@@ -58,12 +58,15 @@ i           Prompt you to enter bytes which will be (i)nserted at current index
 /deadbeef/i If bytes de ad be ef exist after current index, move there
               and prompt you to enter bytes which will be (i)nserted there
 12,3dp      (p)rint bytes 12 - 3d inclusive, move to byte 12
+l           (l)oad a new file.
+L           (L)oad state from a file.  Fails if file you were editing is gone.
 m           Toggle whether or not characters are printed after bytes
 n           Toggle whether or not byte (n)umbers are printed before bytes
 o           Toggle using c(o)lor
 p           (p)rint current line of byte(s) (depending on 'W')
 R           Toggle (R)ead-only mode
 s           Print (s)tate of toggles, 'W'idth, etc.
+S           (S)ave state to a file except the bytes you're editing.
 t3d         Print 0x3d lines of con(t)extual bytes after current line [Default 0]
 T3d         Print 0x3d lines of con(T)extual bytes before current line [Default 0]
 u           (u)pdate filename to write to
@@ -133,7 +136,7 @@ impl Command {
         let re_search_again = Regex::new(r"^ *(?P<direction>[/?]) *$").unwrap();
         let re_search_kill = Regex::new(r"^ */(?P<bytes>[0-9a-fA-F]+)/k *$").unwrap();
         let re_search_insert = Regex::new(r"^ */(?P<bytes>[0-9a-fA-F]+)/i *$").unwrap();
-        let re_single_char_command = Regex::new(r"^ *(?P<command>[hijkmnopqRsuwx]).*$").unwrap();
+        let re_single_char_command = Regex::new(r"^ *(?P<command>[hijkmnopqRsSlLuwx]).*$").unwrap();
         let re_range = Regex::new(r"^ *(?P<begin>[0-9a-fA-F.$]+) *, *(?P<end>[0-9a-fA-F.$]+) *(?P<the_rest>.*) *$").unwrap();
         let re_specified_index = Regex::new(r"^ *(?P<index>[0-9A-Fa-f.$]+) *(?P<the_rest>.*) *$").unwrap();
         let re_offset_index = Regex::new(r"^ *(?P<sign>[-+])(?P<offset>[0-9A-Fa-f]+) *(?P<the_rest>.*) *$").unwrap();
@@ -654,6 +657,81 @@ pub fn update_filename(state: &mut ec::State) {
     let filename = filename.unwrap();
 
     state.filename = filename;
+
+    /* Nothing's been written to that file yet, so */
+    state.unsaved_changes = true;
+}
+
+
+pub fn load_new_file(state: &mut ec::State) {
+	if state.unsaved_changes {
+		let unsaved_prompt = "You have unsaved changes.  Carry on? (y/n): ";
+		println!("{}", unsaved_prompt);
+		let yeses = vec!["y", "Y", "Yes", "yes"];
+		let nos   = vec!["n", "N", "No",  "no"];
+		let carry_on = loop {
+			let carry_on_s = read_string_from_user(Some(""));
+			if carry_on_s.is_err() {
+				println!("? {:?}", carry_on_s);
+				return;
+			}
+			let carry_on_s = carry_on_s.unwrap();
+
+			if yeses.contains(&carry_on_s.as_str()) {
+				break true;
+			}
+			if nos.contains(&carry_on_s.as_str()) {
+				break false;
+			}
+			println!("{}", unsaved_prompt);
+		};
+		if !carry_on {
+			return;
+		}
+	}
+
+    let filename =
+            read_string_from_user(Some(
+                    "Enter filename from which to load bytes: "));
+    if filename.is_err() {
+        println!("? {:?}", filename);
+        return;
+    }
+    let filename = filename.unwrap();
+
+    let maybe_all_bytes =
+            ec::all_bytes_from_filename(&filename);
+    if maybe_all_bytes.is_ok() {
+        state.filename = filename;
+        state.all_bytes = maybe_all_bytes.unwrap();
+        return;
+    }
+
+    match maybe_all_bytes {
+        Err(ec::AllBytesFromFilenameError::NotARegularFile) => {
+            println!("? {} is not a regular file", filename);
+        },
+        Err(ec::AllBytesFromFilenameError::FileDoesNotExist) => {
+            println!("? {} does not exist", filename);
+            println!("Use 'u' to just change filename");
+        },
+        _ => {
+            println!("? {:?}", maybe_all_bytes);
+        },
+    }
+}
+
+
+pub fn save_state(state: &ec::State) {
+    let filename = read_string_from_user(Some("Enter filename to save state: "));
+    if filename.is_ok() {
+        if let Err(error) = state.write_to_disk(&filename.unwrap()) {
+            println!("? {:?}", error);
+        }
+    }
+    else {
+        println!("? {:?}", filename);
+    }
 }
 
 
@@ -717,11 +795,24 @@ pub fn actual_runtime(filename:&str, quiet:bool, color:bool, readonly:bool)
             }
             else {
                 let maybe_all_bytes = ec::all_bytes_from_filename(filename);
-                if maybe_all_bytes.is_err() {
-                    println!("{:?}", maybe_all_bytes);
-                    return 1;
+                if maybe_all_bytes.is_ok() {
+                    maybe_all_bytes.unwrap()
+                }
+                else {
+                    match maybe_all_bytes {
+                        Err(ec::AllBytesFromFilenameError::NotARegularFile) => {
+                            println!("{} is not a regular file", filename);
+                            return 1;
+                        },
+                        Err(ec::AllBytesFromFilenameError::FileDoesNotExist) => {
+                            Vec::new()
+                        },
+                        _ => {
+                            println!("Cannot read {}", filename);
+                            return 1;
+                        }
+                    }
                 }
-                maybe_all_bytes.unwrap()
             }
         ,
         // TODO calculate based on longest possible index
@@ -861,6 +952,31 @@ pub fn actual_runtime(filename:&str, quiet:bool, color:bool, readonly:bool)
                         state.print_bytes();
                     },
 
+
+                    /* Load new file */
+                    'l' => {
+                        load_new_file(&mut state);
+                    }
+
+                    /* Load state from a file */
+                    'L' => {
+                        let filename =
+                                read_string_from_user(Some(
+                                        "Enter filename from which to load state: "));
+                        if filename.is_ok() {
+                            let new_state = State::from_filename(&filename.unwrap());
+                            if new_state.is_ok() {
+                                state = new_state.unwrap();
+                            }
+                            else {
+                                println!("? {:?}", new_state);
+                            }
+                        }
+                        else {
+                            println!("? {:?}", filename);
+                        }
+                    },
+
                     /* Toggle showing char representations of bytes */
                     'm' => {
                         state.show_chars = !state.show_chars;
@@ -976,6 +1092,10 @@ pub fn actual_runtime(filename:&str, quiet:bool, color:bool, readonly:bool)
                         state.readonly = !state.readonly;
                     },
 
+                    /* Write state to a file */
+                    'S' => {
+                        save_state(&state);
+                    },
 
                     /* Print state */
                     's' => {
